/**
 * Restarting simulations
 * 
 * This example demonstrates how to restart a simulation
 * using a binary file. A shearing sheet ring simulation is used, but
 * the same method can be applied to any other type of simulation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "rebound.h"

void heartbeat(struct reb_simulation* const r) ;

float timeout, dt, P, dtout, t1, t2;

#define IAS15PRECISSION 1.e-9
#define IAS15MINDT 0.0

#define	G 1.0
#define	PI 3.14159265358979323844
#define PLANETARYDENSITY 5.5
#define MSOL_SI 1.98855e30
#define AU_SI 149597870700.0
#define RHO2CGS (MSOL_SI*1000.0/pow(AU_SI*100.0,3.0))

#define real double

int ResolveCollisions (rsim, coll)
struct reb_simulation *rsim;
struct reb_collision coll;
{
  int value=0, pi, pj, pk=0;
  struct reb_particle* const particles = rsim->particles;	/* note: only designed for merging planets at this point !!! */
  pi = coll.p1;		/* array indices of particles participating in the collision have no strict order! both i<j and j<i are possible */
  pj = coll.p2;

  FILE *mergers = fopen("nbody.mergers.dat", "a");

    printf ("\n Merger detected. Merging planets %d and %d\n", particles[pi].hash, particles[pj].hash);
    printf ("Previous no. of planets: %d\n", rsim->N_active-1);
    fprintf (mergers, "%.12g\t%d\t%d\n", rsim->t, pi, pj);
    fprintf (mergers, "%d\t%f\t%.8g\n", particles[pi].hash, particles[pi].m, particles[pi].r);
    fprintf (mergers, "%#.18g\t%#.18g\t%#.18g\n", particles[pi].x, particles[pi].y, particles[pi].z);
    fprintf (mergers, "%#.18g\t%#.18g\t%#.18g\n", particles[pi].vx, particles[pi].vy, particles[pi].vz);
    fprintf (mergers, "%d\t%f\t%.8g\n", particles[pj].hash, particles[pj].m, particles[pj].r);
    fprintf (mergers, "%#.18g\t%#.18g\t%#.18g\n", particles[pj].x, particles[pj].y, particles[pj].z);
    fprintf (mergers, "%#.18g\t%#.18g\t%#.18g\n", particles[pj].vx, particles[pj].vy, particles[pj].vz);

  value = reb_collision_resolve_merge (rsim, coll);
  rsim->N_active--;		/* change number of massive bodies accordingly */

    switch (value) {
      case 1 :
        ; pk = pj;	/* i-th particle will be discarded, j-th particle replaced by the merger -> output it */
        break;
      case 2 :
        ; pk = pi;	/* vice versa */
        break;
    }
    fprintf (mergers, "%d\t%f\t%.8g\n", particles[pk].hash, particles[pk].m, particles[pk].r);
    fprintf (mergers, "%#.18g\t%#.18g\t%#.18g\n", particles[pk].x, particles[pk].y, particles[pk].z);
    fprintf (mergers, "%#.18g\t%#.18g\t%#.18g\n", particles[pk].vx, particles[pk].vy, particles[pk].vz);
    printf ("No. of planets after the merger: %d\n", rsim->N_active-1);

    fclose(mergers);

  return value;
}

struct reb_simulation *reb_input_ascii(char* filename){
  
  char s[512], nm[512], *s1;
  int i, err;
  real mass, gas, solid, a, e, inc, Omega, omega, f;
  real rho, radius, rad;

  struct reb_simulation *rsim = reb_create_simulation ();

  struct reb_particle primary = {0};
  primary.hash = 0;
  primary.m = 1.0;
  reb_add (rsim, primary);

  FILE* input = fopen(filename, "r");
  if (input == NULL) {
    fprintf(stderr, "Error : can't find '%s'.\n", filename);
    exit(1);
  }

  rad = PI/180.0;
  rho = PLANETARYDENSITY/RHO2CGS;
  i = 0;
  while (fgets(s, 510, input) != NULL) {
    sscanf(s, "%s ", nm);
    if (!isalpha(s[0])) continue;

    s1 = s + strlen(nm);
    err = sscanf(s1 + strspn(s1, "\t :=>_"), "%lf %lf %lf %lf %lf %lf %lf %lf %lf", \
      &mass, &gas, &solid, &a, &e, &inc, &Omega, &omega, &f);
    if (err != 9) {
      fprintf (stderr, "Error reading file '%s'.\n", filename);
      exit(1);
    }

    err = 0;
    struct reb_particle p = reb_tools_orbit_to_particle_err (G, primary, \
      (real)mass, (real)a, (real)e, (real)inc*rad, (real)Omega*rad, (real)omega*rad, (real)f*rad, &err);
    if (err != 0) {
      printf("Error: Planet no. %d cannot be initialized. Terminating now...\n", i);
      exit(1);
    }
    i++;
    p.hash = i;
    radius = pow(3.0*(real)mass/(4.0*PI*rho), 1.0/3.0);
    p.r = radius;
    p.gas = gas;
    p.solid = solid;
    reb_add (rsim, p);	
  }

  fclose(input);

  rsim->N_active = i + 1;
  return rsim;
}

void reb_output_ascii(struct reb_simulation* r, char* filename){
    const int N = r->N;
#ifdef MPI
    char filename_mpi[1024];
    sprintf(filename_mpi,"%s_%d",filename,r->mpi_id);
    FILE* of = fopen(filename_mpi,"a");
#else // MPI
    FILE* of = fopen(filename,"a");
#endif // MPI
    if (of==NULL){
        reb_exit("Can not open file.");
    }
    for (int i=1; i<N; i++){
        struct reb_particle p = r->particles[i];
        struct reb_orbit orbit = reb_tools_particle_to_orbit (G, r->particles[i], r->particles[0]);
        fprintf(of, "%d\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\t%16.12g\n", p.hash, r->t, orbit.a, orbit.e, orbit.inc, orbit.Omega, orbit.omega, orbit.f, p.m, p.r, p.x, p.y, p.z, p.vx, p.vy, p.vz);
    }
    fclose(of);
}

#define EPS 1.0e-12

void heartbeat(struct reb_simulation* const rsim){
//  printf("rsim->t = %f\n", rsim->t);
  if (rsim->t >= timeout-EPS) {
    printf("timeout = %f\n", timeout);
    reb_output_ascii(rsim, "output.txt");
    timeout += dtout;
  }
}

int main(int argc, char* argv[]){

  printf("Init\n");
//  struct reb_simulation* rsim = reb_create_simulation_from_binary("nbody.simulation.dat");
//  if (rsim == NULL)
//    exit(1);
  struct reb_simulation* rsim = reb_input_ascii("embryos.cfg");
  if (rsim == NULL)
    exit(1);

  printf("Setup\n");
  rsim->integrator = REB_INTEGRATOR_IAS15;
  rsim->ri_ias15.epsilon = IAS15PRECISSION;
  rsim->ri_ias15.min_dt = IAS15MINDT;
  rsim->collision = REB_COLLISION_DIRECT;
  rsim->collision_resolve = ResolveCollisions;
  rsim->heartbeat = heartbeat;

  dt = 0.31415;
  P = 20*dt;
  dtout = 10*P;
  t1 = 0.0;
  t2 = 10000.0*P;

  printf("Ready!\n");
  timeout = rsim->t;
  printf("t1 = %f\n", t1);
  printf("t2 = %f\n", t2);
  printf("t2-t1 = %f\n", t2-t1);
  printf("rsim->t = %f\n", rsim->t);
  printf("timeout = %f\n", timeout);

  reb_integrate(rsim, t2);

  reb_output_binary(rsim, "final.bin");

  printf("Done.\n");
}


